import uuid
from abc import ABC, abstractmethod, abstractproperty
from typing import Union

from django.db.models.query import QuerySet
from django.db.models import Model
from insuree.services import validate_insuree_number
from api_fhir_r4.converters import ReferenceConverterMixin
from django.core.exceptions import ValidationError

class GenericModelRetriever(ABC):

    @property
    @abstractmethod
    def identifier_field(self) -> str:
        pass

    @property
    @abstractmethod
    def serializer_reference_type(self) -> Union[
        ReferenceConverterMixin.UUID_REFERENCE_TYPE,
        ReferenceConverterMixin.CODE_REFERENCE_TYPE,
        ReferenceConverterMixin.DB_ID_REFERENCE_TYPE
    ]:
        pass

    @classmethod
    @abstractmethod
    def identifier_validator(cls, identifier_value) -> bool:
        pass

    @classmethod
    def retriever_additional_queryset_filtering(cls, queryset):
        # By default no additional changes are made in queryset
        return queryset

    @classmethod
    def get_model_object(cls, queryset: QuerySet, identifier_value) -> Model:
        filters = {}
        if cls.serializer_reference_type == 'uuid_reference':
            identifier_value = uuid.UUID(str(identifier_value))
        elif hasattr(queryset.model, 'validity_to'):  
            filters['validity_to__isnull'] = True
        filters[cls.identifier_field] = identifier_value
        try:
            return queryset.get(**filters)
        except Exception as e:
            raise ValidationError(f"failed to retrieve {queryset.model.__name__} with the filter {filters}; details {e}")


class UUIDIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = 'uuid'
    serializer_reference_type = ReferenceConverterMixin.UUID_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return cls._is_uuid_identifier(identifier_value)

    @classmethod
    def _is_uuid_identifier(cls, identifier):
        try:
            uuid.UUID(str(identifier))
            return True
        except ValueError:
            return False


class DatabaseIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = 'id'
    serializer_reference_type = ReferenceConverterMixin.DB_ID_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return isinstance(identifier_value, int) or identifier_value.isdigit()


class CodeIdentifierModelRetriever(GenericModelRetriever):
    identifier_field = 'code'
    serializer_reference_type = ReferenceConverterMixin.CODE_REFERENCE_TYPE

    @classmethod
    def identifier_validator(cls, identifier_value):
        return isinstance(identifier_value, str)

    


class CHFIdentifierModelRetriever(CodeIdentifierModelRetriever):
    identifier_field = 'chf_id'

    @classmethod
    def identifier_validator(cls, identifier_value):
        # From model specification
        return isinstance(identifier_value, str) and validate_insuree_number(identifier_value)


class GroupIdentifierModelRetriever(CHFIdentifierModelRetriever):
    identifier_field = 'head_insuree_id__chf_id'




from drf_spectacular.utils import extend_schema, extend_schema_view
from graphql_jwt.utils import jwt_payload

from api_fhir_r4.openapi_schema_extensions import get_inline_login_request_serializer, \
    get_inline_login_200_response_serializer
from core.jwt import *
from core.models import User
from core.services import user_authentication
from rest_framework import viewsets
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework import exceptions


@extend_schema_view(
    create=extend_schema(
        request=get_inline_login_request_serializer(),
        responses={
            (200, 'application/json'): get_inline_login_200_response_serializer(),
            (400,): None,
            (401,): None
        }
    )
)
class LoginView(viewsets.ViewSet):
    permission_classes = (AllowAny,)

    def create(self, request, *args, **kwargs):
        data = request.data
        # check if we have both required data in request payload

        username = data.get('username')
        password = data.get('password')
        try:
            request.user = user_authentication(request, username, password)
        except exceptions.ParseError as e:
            return Response(str(e), status=400)
        except exceptions.AuthenticationFailed as e:
            return Response(str(e), status=401)
        if request.user:
            # take the payload base on user data - using same mechanism as
            # in graphql_jwt with generating payload.
            payload = jwt_payload(user=request.user)
            # encode token based on payload
            token = jwt_encode_user_key(payload=payload, context=request)
            if token:
                # return ok
                response = {
                    "token": token,
                    "exp": payload['exp'],
                }
                return Response(data=response, status=200)
            # return unauthorized
            return Response(status=401)
        else:
            # return bad request
            return Response(status=400)



import json
import os

from rest_framework import status
from rest_framework.test import APITestCase

from api_fhir_r4.configurations import GeneralConfiguration
from api_fhir_r4.tests import GenericFhirAPITestMixin
from django.utils.translation import gettext as _

from api_fhir_r4.tests.utils import get_connection_payload,get_or_create_user_api

class AuthorizationAPITests(GenericFhirAPITestMixin, APITestCase):
    base_url = GeneralConfiguration.get_base_url()
    url_to_test_authorization = base_url + 'Group/'
    _test_json_path = "/test/test_login.json"
    _test_json_path_credentials = "/test/test_login.json"
    _test_request_data_credentials = None

    def setUp(self):
        super(AuthorizationAPITests, self).setUp()
        self.test_user = get_or_create_user_api()

    def get_bundle_from_json_response(self, response):
        pass

    def test_post_should_authorize_correctly(self):
        response = self.client.post(self.base_url + 'login/', data=get_connection_payload(), format='json')
        response_json = response.json()
        token = response_json["token"]
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        content = json.loads(response.content)
        self.assertEqual(response.status_code, status.HTTP_200_OK, f"{response.content}")

    def test_post_should_raise_no_auth_header(self):
        response = self.client.get(self.url_to_test_authorization, format='json')
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(self.get_response_details(response_json), "Authentication credentials were not provided.")

    def test_post_should_raise_error_decode_token(self):
        response = self.client.post(self.base_url + 'login/', data=get_connection_payload(), format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}ssdd"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(self.get_response_details(response_json), _('INCORRECT_CREDENTIALS'))

    def test_post_should_raise_lack_of_bearer_prefix(self):
        response = self.client.post(self.base_url + 'login/', data=get_connection_payload(), format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"{token}"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        self.assertTrue(self.get_response_details(response_json) != '',
                        msg="401 Response without error message")

    def test_post_should_raise_unproper_structure_of_token(self):
        response = self.client.post(self.base_url + 'login/', data=get_connection_payload(), format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token} xxxxx xxxxxx"
        }
        response = self.client.get(self.url_to_test_authorization, format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertTrue(self.get_response_details(response_json) != '',
                        msg="401 Response without error message")

    def test_post_should_raise_forbidden(self):
        _TEST_DATA_USER_NO_ROLE = {
            "username": "TestUserTest3",
            "last_name": "TestUserTest3",
            "password": "TestPasswordTest3",
            "other_names": "TestUserTest3",
            "user_types": "INTERACTIVE",
            "language": "en",
            "roles": [1],
        }
        get_or_create_user_api(_TEST_DATA_USER_NO_ROLE)
        response = self.client.post(self.base_url + 'login/', data=get_connection_payload(_TEST_DATA_USER_NO_ROLE), format='json')
        response_json = response.json()
        token = response_json["token"]
        headers = {
            "Content-Type": "application/json",
            'HTTP_AUTHORIZATION': f"Bearer {token}"
        }
        response = self.client.get(self.base_url + 'Claim/', format='json', **headers)
        response_json = response.json()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(
            self.get_response_details(response_json), "You do not have permission to perform this action."
        )

    def test_get_should_required_login(self):
        pass



from core import datetime
from claim.models import Claim, ClaimItem, ClaimService

from insuree.test_helpers import create_test_insuree
from location.models import HealthFacility
from medical.test_helpers import create_test_item, create_test_service
from api_fhir_r4.configurations import R4CommunicationRequestConfig as Config
from api_fhir_r4.tests import GenericTestMixin, LocationTestMixin
from api_fhir_r4.utils import TimeUtils
from claim.test_helpers import create_test_claim_context
from location.test_helpers import create_test_village, create_test_health_facility
from medical.models import Diagnosis


class CommunicationRequestTestMixin(GenericTestMixin):
    _TEST_CODE = 'codeTest'
    _TEST_UUID = "7ac646cb-d3cd-4660-baeb-ee34ecf0354e"
    _TEST_STATUS = Claim.STATUS_ENTERED
    _TEST_STATUS_DISPLAY = "entered"
    _TEST_OUTCOME = "queued"
    _TEST_ADJUSTMENT = "adjustment"
    _TEST_DATE_PROCESSED = "2010-11-16T00:00:00"
    _TEST_APPROVED = 1000.00
    _TEST_REJECTION_REASON = 0
    _TEST_VISIT_TYPE = "O"

    # claim item data
    _TEST_ITEM_CODE = "iCode"
    _TEST_ITEM_UUID = "e2bc1546-390b-4d41-8571-632ecf7a0936"
    _TEST_ITEM_STATUS = Claim.STATUS_ENTERED
    _TEST_ITEM_QUANTITY = 20
    _TEST_ITEM_PRICE = 10.0
    _TEST_ITEM_REJECTED_REASON = 0

    # claim service data
    _TEST_SERVICE_CODE = "sCode"
    _TEST_SERVICE_UUID = "a17602f4-e9ff-4f42-a6a4-ccefcb23b4d6"
    _TEST_SERVICE_STATUS = Claim.STATUS_ENTERED
    _TEST_SERVICE_QUANTITY = 1
    _TEST_SERVICE_PRICE = 800
    _TEST_SERVICE_REJECTED_REASON = 0

    _TEST_ID = 9999
    _PRICE_ASKED = 1000
    _PRICE_APPROVED = 1000
    _ADMIN_AUDIT_USER_ID = 1

    _TEST_ITEM_AVAILABILITY = True

    _TEST_ITEM_TYPE = 'D'
    _TEST_SERVICE_TYPE = 'D'

    # insuree and claim admin data
    _TEST_INSUREE_UUID = "76aca309-f8cf-4890-8f2e-b416d78de00b"
    _TEST_CLAIM_ADMIN_UUID = "044c33d1-dbf3-4d6a-9924-3797b461e535"

    # hf test data
    _TEST_HF_ID = 10000
    _TEST_HF_UUID = "6d0eea8c-62eb-11ea-94d6-c36229a16c2f"
    _TEST_HF_CODE = "12345678"
    _TEST_HF_NAME = "TEST_NAME"
    _TEST_HF_LEVEL = "H"
    _TEST_HF_LEGAL_FORM = "G"
    _TEST_ADDRESS = "TEST_ADDRESS"
    _TEST_PHONE = "133-996-476"
    _TEST_FAX = "1-408-999 8888"
    _TEST_EMAIL = "TEST@TEST.com"
    test_claim_admin = None
    test_insuree = None
    test_claim = None
    test_hf = None
    test_village = None
    sub_str = {}
    def setUp(self):
        super(CommunicationRequestTestMixin, self).setUp()
        service_props = {
            'status': self._TEST_SERVICE_STATUS,
            'qty_approved': self._TEST_SERVICE_QUANTITY,
            'qty_provided': self._TEST_SERVICE_QUANTITY,
            'rejection_reason': self._TEST_SERVICE_REJECTED_REASON,
            'availability': self._TEST_ITEM_AVAILABILITY,
            'price_asked': self._TEST_SERVICE_PRICE,
            'price_approved': self._TEST_SERVICE_PRICE,
            'audit_user_id': self._ADMIN_AUDIT_USER_ID,
        }
        item_props = {
            'status': self._TEST_ITEM_STATUS,
            'qty_approved': self._TEST_ITEM_QUANTITY,
            'qty_provided': self._TEST_ITEM_QUANTITY,
            'rejection_reason': self._TEST_ITEM_REJECTED_REASON,
            'availability': self._TEST_ITEM_AVAILABILITY,
            'price_asked': self._TEST_ITEM_PRICE,
            'price_approved': self._TEST_ITEM_PRICE,
            'audit_user_id': self._ADMIN_AUDIT_USER_ID,
        }
        
        
        self.test_claim, self.test_insuree, policy, self.test_hf = create_test_claim_context(
            claim={
                'visit_type': self._TEST_VISIT_TYPE,
                'status': self._TEST_STATUS,
                'feedback_status': Claim.FEEDBACK_SELECTED
                }, 
            claim_admin={
                'uuid': self._TEST_CLAIM_ADMIN_UUID,
            },
            insuree={'uuid': self._TEST_INSUREE_UUID}, 
            product={}, 
            hf={
                'uuid': self._TEST_HF_UUID,
                'name': self._TEST_HF_NAME,
                'level': self._TEST_HF_LEVEL,
                'legal_form_id': self._TEST_HF_LEGAL_FORM,
                'address': self._TEST_ADDRESS,
                'phone': self._TEST_PHONE,
                'fax': self._TEST_FAX,
                'email': self._TEST_EMAIL,
            }, 
            items=[
                item_props
                ], 
            services=[
                service_props
            ])        
 


        self.sub_str[self._TEST_HF_UUID]=self.test_hf.uuid
        self.sub_str[self._TEST_CLAIM_ADMIN_UUID]=self.test_claim.admin.uuid
        self.sub_str[self._TEST_INSUREE_UUID]=self.test_insuree.uuid
        self.sub_str[self._TEST_UUID]=self.test_claim.uuid
        self.sub_str[self._TEST_SERVICE_UUID]=self.test_claim.services.first().service.uuid
        self.sub_str[self._TEST_ITEM_UUID]=self.test_claim.items.first().item.uuid

   

    def create_test_imis_instance(self):
        
        return self.test_claim

    def verify_fhir_instance(self, fhir_obj):
        self.assertEqual("active", fhir_obj.status)
        self.assertEqual(f"{Claim.FEEDBACK_SELECTED}", fhir_obj.statusReason.coding[0].code)
        self.assertIn(str(self.test_insuree.uuid), fhir_obj.subject.reference)
        self.assertIn(str(self.test_claim.admin.uuid), fhir_obj.recipient[0].reference)
        self.assertIn(str(self.test_claim.uuid), fhir_obj.about[0].reference)
        for payload in fhir_obj.payload:
            code = payload.extension[0].valueCodeableConcept.coding[0].code
            content_string = payload.contentString
            if code == Config.get_fhir_care_rendered_code():
                self.assertEqual("Care Rendered? (yes|no)", content_string)
            elif code == Config.get_fhir_payment_asked_code():
                self.assertEqual("Payment Asked? (yes|no)", content_string)
            elif code == Config.get_fhir_drug_prescribed_code():
                self.assertEqual("Drug Prescribed? (yes|no)", content_string)
            elif code == Config.get_fhir_drug_received_code():
                self.assertEqual("Drug Received? (yes|no)", content_string)
            elif code == Config.get_fhir_asessment_code():
                self.assertEqual("Asessment? (0|1|2|3|4|5)", content_string)
